@@ -1,7 +1,7 @@
 '''
 mrtparse.py - MRT format data parser
 
-Copyright (C) 2016 greenHippo, LLC.
+Copyright (C) 2017 Tetsumune KISO
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -30,7 +30,7 @@ import collections
 import signal
 signal.signal(signal.SIGPIPE, signal.SIG_DFL)
 
-__version__ = '1.5'
+__version__ = '1.6'
 
 # Magic Number
 GZIP_MAGIC = b'\x1f\x8b'
@@ -124,6 +124,12 @@ TD_V2_ST = reverse_defaultdict({
     4:'RIB_IPV6_UNICAST',
     5:'RIB_IPV6_MULTICAST',
     6:'RIB_GENERIC',
+    7:'GEO_PEER_TABLE',              # Defined in RFC6397
+    8:'RIB_IPV4_UNICAST_ADDPATH',    # Defined in RFC8050
+    9:'RIB_IPV4_MULTICAST_ADDPATH',  # Defined in RFC8050
+    10:'RIB_IPV6_UNICAST_ADDPATH',   # Defined in RFC8050
+    11:'RIB_IPV6_MULTICAST_ADDPATH', # Defined in RFC8050
+    12:'RIB_GENERIC_ADDPATH',        # Defined in RFC8050
 })
 
 # BGP4MP,BGP4MP_ET Subtypes
@@ -131,12 +137,16 @@ TD_V2_ST = reverse_defaultdict({
 BGP4MP_ST = reverse_defaultdict({
     0:'BGP4MP_STATE_CHANGE',
     1:'BGP4MP_MESSAGE',
-    2:'BGP4MP_ENTRY',             # Deprecated in RFC6396
-    3:'BGP4MP_SNAPSHOT',          # Deprecated in RFC6396
+    2:'BGP4MP_ENTRY',                      # Deprecated in RFC6396
+    3:'BGP4MP_SNAPSHOT',                   # Deprecated in RFC6396
     4:'BGP4MP_MESSAGE_AS4',
     5:'BGP4MP_STATE_CHANGE_AS4',
     6:'BGP4MP_MESSAGE_LOCAL',
     7:'BGP4MP_MESSAGE_AS4_LOCAL',
+    8:'BGP4MP_MESSAGE_ADDPATH',            # Defined in RFC8050
+    9:'BGP4MP_MESSAGE_AS4_ADDPATH',        # Defined in RFC8050
+    10:'BGP4MP_MESSAGE_LOCAL_ADDPATH',     # Defined in RFC8050
+    11:'BGP4MP_MESSAGE_AS4_LOCAL_ADDPATH', # Defined in RFC8050
 })
 
 # MRT Message Subtypes
@@ -397,6 +407,29 @@ def as_repr(n=None):
     except AttributeError:
         return AS_REPR['asplain']
 
+def af_num(afi=None, safi=None):
+    '''
+    the values of AFI/SAFI.
+    '''
+    if afi is not None:
+        af_num.afi = afi
+        af_num.safi = safi
+    try:
+        return (af_num.afi, af_num.safi)
+    except AttributeError:
+        return (0, 0)
+
+def is_add_path(f=None):
+    '''
+    Flag for add-path.
+    '''
+    if f is not None:
+        is_add_path.f = f
+    try:
+        return is_add_path.f
+    except AttributeError:
+        return False
+
 class MrtFormatError(Exception):
     '''
     Exception for invalid MRT formatted data.
@@ -412,7 +445,8 @@ class Base:
     __slots__ = ['buf', 'p']
 
     def __init__(self):
-        self.buf = None
+        for slot in self.__slots__:
+            setattr(self, slot, None)
         self.p = 0
 
     def chk_buf(self, n):
@@ -421,7 +455,7 @@ class Base:
         '''
         if len(self.buf) - self.p < n:
             raise MrtFormatError(
-                'Insufficient buffer %d < %d'
+                'Insufficient buffer %d < %d byte'
                 % (len(self.buf) - self.p, n))
 
     def val_num(self, n):
@@ -506,6 +540,8 @@ class Base:
         Convert buffers to NLRI.
         '''
         try:
+            if is_add_path():
+                raise MrtFormatError
             p = self.p
             l = []
             while p < n:
@@ -528,12 +564,10 @@ class Reader(Base):
     '''
     Reader for MRT format data.
     '''
-    __slots__ = ['mrt', 'f']
+    __slots__ = ['buf', 'p', 'mrt', 'f']
 
     def __init__(self, arg):
         Base.__init__(self)
-        self.mrt = None
-        self.f = None
 
         # for file instance
         if hasattr(arg, 'read'):
@@ -588,12 +622,14 @@ class Reader(Base):
         Decoder for MRT header.
         '''
         as_len(4)
+        af_num(0, 0)
+        is_add_path(False)
         self.mrt = Mrt(self.f.read(12))
         if len(self.mrt.buf) == 0:
             self.close()
         elif len(self.mrt.buf) < 12:
             raise MrtFormatError(
-                'Invalid MRT header length %d < 12'
+                'Invalid MRT header length %d < 12 byte'
                 % len(self.mrt.buf))
         self.mrt.unpack()
 
@@ -605,9 +641,16 @@ class Reader(Base):
         self.mrt.buf += data
         if len(data) < self.mrt.len:
             raise MrtFormatError(
-                'Invalid MRT data length %d < %d'
+                'Invalid MRT data length %d < %d byte'
                 % (len(data), self.mrt.len))
 
+        if len(MRT_ST[self.mrt.type]) \
+            and MRT_ST[self.mrt.type][self.mrt.subtype] == 'Unknown':
+            raise MrtFormatError(
+                'Unsupported %s subtype %d(%s)'
+                % (self.mrt.type, self.mrt.subtype,
+                MRT_ST[self.mrt.type][self.mrt.subtype]))
+
         if self.mrt.type == MRT_T['TABLE_DUMP_V2']:
             self.unpack_td_v2(data)
         elif self.mrt.type == MRT_T['BGP4MP'] \
@@ -639,18 +682,31 @@ class Reader(Base):
         '''
         Decoder for Table_Dump_V2 format.
         '''
+        if self.mrt.subtype == TD_V2_ST['RIB_IPV4_UNICAST_ADDPATH'] \
+            or self.mrt.subtype == TD_V2_ST['RIB_IPV4_MULTICAST_ADDPATH'] \
+            or self.mrt.subtype == TD_V2_ST['RIB_IPV6_UNICAST_ADDPATH'] \
+            or self.mrt.subtype == TD_V2_ST['RIB_IPV6_MULTICAST_ADDPATH']:
+            is_add_path(True)
+
         if self.mrt.subtype == TD_V2_ST['RIB_IPV4_UNICAST'] \
-            or self.mrt.subtype == TD_V2_ST['RIB_IPV4_MULTICAST']:
+            or self.mrt.subtype == TD_V2_ST['RIB_IPV4_MULTICAST'] \
+            or self.mrt.subtype == TD_V2_ST['RIB_IPV4_UNICAST_ADDPATH'] \
+            or self.mrt.subtype == TD_V2_ST['RIB_IPV4_MULTICAST_ADDPATH']:
+            af_num.afi = AFI_T['IPv4']
             self.mrt.rib = AfiSpecRib(data)
-            self.mrt.rib.unpack(AFI_T['IPv4'])
+            self.mrt.rib.unpack()
         elif self.mrt.subtype == TD_V2_ST['RIB_IPV6_UNICAST'] \
-            or self.mrt.subtype == TD_V2_ST['RIB_IPV6_MULTICAST']:
+            or self.mrt.subtype == TD_V2_ST['RIB_IPV6_MULTICAST'] \
+            or self.mrt.subtype == TD_V2_ST['RIB_IPV6_UNICAST_ADDPATH'] \
+            or self.mrt.subtype == TD_V2_ST['RIB_IPV6_MULTICAST_ADDPATH']:
+            af_num.afi = AFI_T['IPv6']
             self.mrt.rib = AfiSpecRib(data)
-            self.mrt.rib.unpack(AFI_T['IPv6'])
+            self.mrt.rib.unpack()
         elif self.mrt.subtype == TD_V2_ST['PEER_INDEX_TABLE']:
             self.mrt.peer = PeerIndexTable(data)
             self.mrt.peer.unpack()
-        elif self.mrt.subtype == TD_V2_ST['RIB_GENERIC']:
+        elif self.mrt.subtype == TD_V2_ST['RIB_GENERIC'] \
+            or self.mrt.subtype == TD_V2_ST['RIB_GENERIC_ADDPATH']:
             self.mrt.rib = RibGeneric(data)
             self.mrt.rib.unpack()
         else:
@@ -661,24 +717,13 @@ class Mrt(Base):
     Class for MRT header.
     '''
     __slots__ = [
-        'ts', 'type', 'subtype', 'len', 'micro_ts', 'bgp', 'peer', 'td', 'rib',
-        'err', 'err_msg'
+        'buf', 'p', 'ts', 'type', 'subtype', 'len', 'micro_ts', 'bgp', 'peer',
+        'td', 'rib', 'err', 'err_msg'
     ]
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.ts = None
-        self.type = None
-        self.subtype = None
-        self.len = None
-        self.micro_ts = None
-        self.bgp = None
-        self.peer = None
-        self.td = None
-        self.rib = None
-        self.err = None
-        self.err_msg = None
 
     def unpack(self):
         '''
@@ -695,23 +740,13 @@ class TableDump(Base):
     Class for Table_Dump format.
     '''
     __slots__ = [
-        'view', 'seq', 'prefix', 'plen', 'status', 'org_time', 'peer_ip',
-        'peer_as', 'attr_len', 'attr'
+        'buf', 'p', 'view', 'seq', 'prefix', 'plen', 'status', 'org_time',
+        'peer_ip', 'peer_as', 'attr_len', 'attr'
     ]
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.view = None
-        self.seq = None
-        self.prefix = None
-        self.plen = None
-        self.status = None
-        self.org_time = None
-        self.peer_ip = None
-        self.peer_as = None
-        self.attr_len = None
-        self.attr = None
 
     def unpack(self, subtype):
         '''
@@ -723,10 +758,14 @@ class TableDump(Base):
         self.plen = self.val_num(1)
         self.status = self.val_num(1)
         self.org_time = self.val_num(4)
+
+        # Considering the IPv4 peers advertising IPv6 Prefixes, first,
+        # the Peer IP Address field is decoded as an IPv4 address.
         self.peer_ip = self.val_addr(AFI_T['IPv4'])
         if subtype == AFI_T['IPv6'] and self.val_num(12):
             self.p -= 16
             self.peer_ip = self.val_addr(subtype)
+
         self.peer_as = self.val_asn(as_len(2))
         attr_len = self.attr_len = self.val_num(2)
         self.attr = []
@@ -741,16 +780,11 @@ class PeerIndexTable(Base):
     '''
     Class for PEER_INDEX_TABLE format.
     '''
-    __slots__ = ['collector', 'view_len', 'view', 'count', 'entry']
+    __slots__ = ['buf', 'p', 'collector', 'view_len', 'view', 'count', 'entry']
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.collector = None
-        self.view_len = None
-        self.view = None
-        self.count = None
-        self.entry = None
 
     def unpack(self):
         '''
@@ -771,15 +805,11 @@ class PeerEntries(Base):
     '''
     Class for Peer Entries.
     '''
-    __slots__ = ['type', 'bgp_id', 'ip', 'asn']
+    __slots__ = ['buf', 'p', 'type', 'bgp_id', 'ip', 'asn']
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.type = None
-        self.bgp_id = None
-        self.ip = None
-        self.asn = None
 
     def unpack(self):
         '''
@@ -787,46 +817,36 @@ class PeerEntries(Base):
         '''
         self.type = self.val_num(1)
         self.bgp_id = self.val_addr(AFI_T['IPv4'])
-        if self.type & 0x01:
-            af = AFI_T['IPv6']
-        else:
-            af = AFI_T['IPv4']
-        self.ip = self.val_addr(af)
-        n = 4 if self.type & (0x01 << 1) else 2
-        self.asn = self.val_asn(n)
+        af_num.afi = AFI_T['IPv6'] if self.type & 0x01 else AFI_T['IPv4']
+        self.ip = self.val_addr(af_num.afi)
+        self.asn = self.val_asn(4 if self.type & (0x01 << 1) else 2)
         return self.p
 
 class RibGeneric(Base):
     '''
     Class for RIB_GENERIC format.
     '''
-    __slots__ = ['seq', 'afi', 'safi', 'nlri', 'count', 'entry']
+    __slots__ = ['buf', 'p', 'seq', 'afi', 'safi', 'nlri', 'count', 'entry']
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.seq = None
-        self.afi = None
-        self.safi = None
-        self.nlri = None
-        self.count = None
-        self.entry = None
 
     def unpack(self):
         '''
         Decoder for RIB_GENERIC format.
         '''
         self.seq = self.val_num(4)
-        self.afi = self.val_num(2)
-        self.safi = self.val_num(1)
+        af_num.afi = self.afi = self.val_num(2)
+        af_num.safi = self.safi = self.val_num(1)
         n = self.val_num(1)
         self.p -= 1
-        self.nlri = self.val_nlri(self.p + (n + 7) // 8, self.afi, self.safi)
+        self.nlri = self.val_nlri(self.p+(n+7)//8, self.afi, self.safi)
         self.count = self.val_num(2)
         self.entry = []
         for _ in range(self.count):
             entry = RibEntries(self.buf[self.p:])
-            self.p += entry.unpack(self.afi)
+            self.p += entry.unpack()
             self.entry.append(entry)
         return self.p
 
@@ -834,29 +854,24 @@ class AfiSpecRib(Base):
     '''
     Class for AFI/SAFI-Specific RIB format.
     '''
-    __slots__ = ['seq', 'plen', 'prefix', 'count', 'entry']
+    __slots__ = ['buf', 'p', 'seq', 'plen', 'prefix', 'count', 'entry']
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.seq = None
-        self.plen = None
-        self.prefix = None
-        self.count = None
-        self.entry = None
 
-    def unpack(self, af):
+    def unpack(self):
         '''
         Decoder for AFI/SAFI-Specific RIB format.
         '''
         self.seq = self.val_num(4)
         self.plen = self.val_num(1)
-        self.prefix = self.val_addr(af, self.plen)
+        self.prefix = self.val_addr(af_num.afi, self.plen)
         self.count = self.val_num(2)
         self.entry = []
         for _ in range(self.count):
             entry = RibEntries(self.buf[self.p:])
-            self.p += entry.unpack(af)
+            self.p += entry.unpack()
             self.entry.append(entry)
         return self.p
 
@@ -864,27 +879,32 @@ class RibEntries(Base):
     '''
     Class for Rib Entries format.
     '''
-    __slots__ = ['peer_index', 'org_time', 'attr_len', 'attr']
+    __slots__ = [
+        'buf', 'p', 'peer_index', 'org_time', 'path_id', 'attr_len', 'attr'
+    ]
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
         self.peer_index = None
         self.org_time = None
+        self.path_id = None
         self.attr_len = None
         self.attr = None
 
-    def unpack(self, af):
+    def unpack(self):
         '''
         Decoder for Rib Entries format.
         '''
         self.peer_index = self.val_num(2)
         self.org_time = self.val_num(4)
+        if is_add_path():
+            self.path_id = self.val_num(4)
         attr_len = self.attr_len = self.val_num(2)
         self.attr = []
         while attr_len > 0:
             attr = BgpAttr(self.buf[self.p:])
-            self.p += attr.unpack(af)
+            self.p += attr.unpack()
             self.attr.append(attr)
             attr_len -= attr.p
         return self.p
@@ -894,22 +914,13 @@ class Bgp4Mp(Base):
     Class for BGP4MP format.
     '''
     __slots__ = [
-        'peer_as', 'local_as', 'ifindex', 'af', 'peer_ip', 'local_ip',
-        'old_state', 'new_state', 'msg'
+        'buf', 'p', 'peer_as', 'local_as', 'ifindex', 'af', 'peer_ip',
+        'local_ip', 'old_state', 'new_state', 'msg'
     ]
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.peer_as = None
-        self.local_as = None
-        self.ifindex = None
-        self.af = None
-        self.peer_ip = None
-        self.local_ip = None
-        self.old_state = None
-        self.new_state = None
-        self.msg = None
 
     def unpack(self, subtype):
         '''
@@ -917,13 +928,21 @@ class Bgp4Mp(Base):
         '''
         if subtype == BGP4MP_ST['BGP4MP_STATE_CHANGE'] \
             or subtype == BGP4MP_ST['BGP4MP_MESSAGE'] \
-            or subtype == BGP4MP_ST['BGP4MP_MESSAGE_LOCAL']:
+            or subtype == BGP4MP_ST['BGP4MP_MESSAGE_LOCAL'] \
+            or subtype == BGP4MP_ST['BGP4MP_MESSAGE_ADDPATH'] \
+            or subtype == BGP4MP_ST['BGP4MP_MESSAGE_LOCAL_ADDPATH']:
             as_len(2)
 
+        if subtype == BGP4MP_ST['BGP4MP_MESSAGE_ADDPATH'] \
+            or subtype == BGP4MP_ST['BGP4MP_MESSAGE_AS4_ADDPATH'] \
+            or subtype == BGP4MP_ST['BGP4MP_MESSAGE_LOCAL_ADDPATH'] \
+            or subtype == BGP4MP_ST['BGP4MP_MESSAGE_AS4_LOCAL_ADDPATH']:
+            is_add_path(True)
+
         self.peer_as = self.val_asn(as_len())
         self.local_as = self.val_asn(as_len())
         self.ifindex = self.val_num(2)
-        self.af = self.val_num(2)
+        af_num.afi = self.af = self.val_num(2)
         self.peer_ip = self.val_addr(self.af)
         self.local_ip = self.val_addr(self.af)
 
@@ -933,7 +952,7 @@ class Bgp4Mp(Base):
             self.new_state = self.val_num(2)
         else:
             self.msg = BgpMessage(self.buf[self.p:])
-            self.p += self.msg.unpack(self.af)
+            self.p += self.msg.unpack()
         return self.p
 
 class BgpMessage(Base):
@@ -941,36 +960,16 @@ class BgpMessage(Base):
     Class for BGP Message.
     '''
     __slots__ = [
-        'marker', 'len', 'type', 'ver', 'my_as', 'holdtime', 'bgp_id',
-        'opt_len', 'opt_params', 'wd_len', 'withdrawn', 'attr_len', 'attr',
-        'nlri', 'err_code', 'err_subcode', 'data', 'afi', 'rsvd', 'safi'
+        'buf', 'p', 'marker', 'len', 'type', 'ver', 'my_as', 'holdtime',
+        'bgp_id', 'opt_len', 'opt_params', 'wd_len', 'withdrawn', 'attr_len',
+        'attr', 'nlri', 'err_code', 'err_subcode', 'data', 'afi', 'rsvd', 'safi'
     ]
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.marker = None
-        self.len = None
-        self.type = None
-        self.ver = None
-        self.my_as = None
-        self.holdtime = None
-        self.bgp_id = None
-        self.opt_len = None
-        self.opt_params = None
-        self.wd_len = None
-        self.withdrawn = None
-        self.attr_len = None
-        self.attr = None
-        self.nlri = None
-        self.err_code = None
-        self.err_subcode = None
-        self.data = None
-        self.afi = None
-        self.rsvd = None
-        self.safi = None
 
-    def unpack(self, af):
+    def unpack(self):
         '''
         Decoder for BGP Message.
         '''
@@ -981,7 +980,7 @@ class BgpMessage(Base):
         if self.type == BGP_MSG_T['OPEN']:
             self.unpack_open()
         elif self.type == BGP_MSG_T['UPDATE']:
-            self.unpack_update(af)
+            self.unpack_update()
         elif self.type == BGP_MSG_T['NOTIFICATION']:
             self.unpack_notification()
         elif self.type == BGP_MSG_T['ROUTE-REFRESH']:
@@ -1006,12 +1005,12 @@ class BgpMessage(Base):
             self.opt_params.append(opt_params)
             opt_len -= opt_params.p
 
-    def unpack_update(self, af):
+    def unpack_update(self):
         '''
         Decoder for BGP UPDATE Message.
         '''
         self.wd_len = self.val_num(2)
-        self.withdrawn = self.val_nlri(self.p + self.wd_len, af)
+        self.withdrawn = self.val_nlri(self.p+self.wd_len, af_num.afi)
         self.attr_len = self.val_num(2)
         attr_len = self.p + self.attr_len
         self.attr = []
@@ -1019,7 +1018,7 @@ class BgpMessage(Base):
             attr = BgpAttr(self.buf[self.p:])
             self.p += attr.unpack()
             self.attr.append(attr)
-        self.nlri = self.val_nlri(self.len, af)
+        self.nlri = self.val_nlri(self.len, af_num.afi)
 
     def unpack_notification(self):
         '''
@@ -1042,22 +1041,13 @@ class OptParams(Base):
     Class for BGP OPEN Optional Parameters.
     '''
     __slots__ = [
-        'type', 'len', 'cap_type', 'cap_len', 'multi_ext', 'orf',
+        'buf', 'p', 'type', 'len', 'cap_type', 'cap_len', 'multi_ext', 'orf',
         'graceful_restart', 'support_as4', 'add_path'
     ]
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.type = None
-        self.len = None
-        self.cap_type = None
-        self.cap_len = None
-        self.multi_ext = None
-        self.orf = None
-        self.graceful_restart = None
-        self.support_as4 = None
-        self.add_path = None
 
     def unpack(self):
         '''
@@ -1161,8 +1151,8 @@ class BgpAttr(Base):
     Class for BGP path attributes
     '''
     __slots__ = [
-        'flag', 'type', 'len', 'origin', 'as_path', 'next_hop', 'med',
-        'local_pref', 'aggr', 'comm', 'org_id', 'cl_list', 'mp_reach',
+        'buf', 'p', 'flag', 'type', 'len', 'origin', 'as_path', 'next_hop',
+        'med', 'local_pref', 'aggr', 'comm', 'org_id', 'cl_list', 'mp_reach',
         'mp_unreach', 'ext_comm', 'as4_path', 'as4_aggr', 'aigp', 'attr_set',
         'large_comm', 'val'
     ]
@@ -1170,29 +1160,8 @@ class BgpAttr(Base):
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.flag = None
-        self.type = None
-        self.len = None
-        self.origin = None
-        self.as_path = None
-        self.next_hop = None
-        self.med = None
-        self.local_pref = None
-        self.aggr = None
-        self.comm = None
-        self.org_id = None
-        self.cl_list = None
-        self.mp_reach = None
-        self.mp_unreach = None
-        self.ext_comm = None
-        self.as4_path = None
-        self.as4_aggr = None
-        self.aigp = None
-        self.attr_set = None
-        self.larg_comm = None
-        self.val = None
 
-    def unpack(self, af=0):
+    def unpack(self):
         '''
         Decoder for BGP path attributes
         '''
@@ -1223,7 +1192,7 @@ class BgpAttr(Base):
         elif self.type == BGP_ATTR_T['CLUSTER_LIST']:
             self.unpack_cluster_list()
         elif self.type == BGP_ATTR_T['MP_REACH_NLRI']:
-            self.unpack_mp_reach_nlri(af)
+            self.unpack_mp_reach_nlri()
         elif self.type == BGP_ATTR_T['MP_UNREACH_NLRI']:
             self.unpack_mp_unreach_nlri()
         elif self.type == BGP_ATTR_T['EXTENDED_COMMUNITIES']:
@@ -1323,7 +1292,7 @@ class BgpAttr(Base):
         while self.p < attr_len:
             self.cl_list.append(self.val_addr(AFI_T['IPv4']))
 
-    def unpack_mp_reach_nlri(self, af):
+    def unpack_mp_reach_nlri(self):
         '''
         Decoder for MP_REACH_NLRI attribute
         '''
@@ -1332,23 +1301,23 @@ class BgpAttr(Base):
         self.mp_reach['afi'] = self.val_num(2)
 
         if AFI_T[self.mp_reach['afi']] != 'Unknown':
-            af = self.mp_reach['afi']
-            self.mp_reach['safi'] = self.val_num(1)
+            af_num.afi = self.mp_reach['afi']
+            af_num.safi = self.mp_reach['safi'] = self.val_num(1)
             self.mp_reach['nlen'] = self.val_num(1)
 
-            if af != AFI_T['IPv4'] and af != AFI_T['IPv6']:
+            if af_num.afi != AFI_T['IPv4'] and af_num.afi != AFI_T['IPv6']:
                 self.p = attr_len
                 return
 
-            if self.mp_reach['safi'] != SAFI_T['UNICAST'] \
-                and self.mp_reach['safi'] != SAFI_T['MULTICAST'] \
-                and self.mp_reach['safi'] != SAFI_T['L3VPN_UNICAST'] \
-                and self.mp_reach['safi'] != SAFI_T['L3VPN_MULTICAST']:
+            if af_num.safi != SAFI_T['UNICAST'] \
+                and af_num.safi != SAFI_T['MULTICAST'] \
+                and af_num.safi != SAFI_T['L3VPN_UNICAST'] \
+                and af_num.safi != SAFI_T['L3VPN_MULTICAST']:
                 self.p = attr_len
                 return
 
-            if self.mp_reach['safi'] == SAFI_T['L3VPN_UNICAST'] \
-                or self.mp_reach['safi'] == SAFI_T['L3VPN_MULTICAST']:
+            if af_num.safi == SAFI_T['L3VPN_UNICAST'] \
+                or af_num.safi == SAFI_T['L3VPN_MULTICAST']:
                 self.mp_reach['rd'] = self.val_rd()
         else:
             self.p -= 2
@@ -1356,14 +1325,14 @@ class BgpAttr(Base):
             self.mp_reach['nlen'] = self.val_num(1)
 
         self.mp_reach['next_hop'] = []
-        self.mp_reach['next_hop'].append(self.val_addr(af))
-        if self.mp_reach['nlen'] == 32 and af == AFI_T['IPv6']:
-            self.mp_reach['next_hop'].append(self.val_addr(af))
+        self.mp_reach['next_hop'].append(self.val_addr(af_num.afi))
+        if self.mp_reach['nlen'] == 32 and af_num.afi == AFI_T['IPv6']:
+            self.mp_reach['next_hop'].append(self.val_addr(af_num.afi))
 
         if 'afi' in self.mp_reach:
             self.mp_reach['rsvd'] = self.val_num(1)
             self.mp_reach['nlri'] = self.val_nlri(
-                attr_len, af, self.mp_reach['safi'])
+                attr_len, af_num.afi, af_num.safi)
 
     def unpack_mp_unreach_nlri(self):
         '''
@@ -1467,16 +1436,11 @@ class Nlri(Base):
     '''
     Class for NLRI.
     '''
-    __slots__ = ['path_id', 'label', 'rd', 'plen', 'prefix']
+    __slots__ = ['buf', 'p', 'path_id', 'label', 'rd', 'plen', 'prefix']
 
     def __init__(self, buf):
         Base.__init__(self)
         self.buf = buf
-        self.path_id = None
-        self.label = None
-        self.rd = None
-        self.plen = None
-        self.prefix = None
 
     def unpack(self, af, saf=0, add_path=0):
         '''
